package com.steerinfo.dil.service.impl;

import com.steerinfo.dil.mapper.AmsContractTruckPriceMapper;
import com.steerinfo.dil.model.AmsContractTransportPrice;
import com.steerinfo.dil.service.IAmsContractTruckPriceService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * AmsContractTransportPrice服务实现：
 * @author generator
 * @version 1.0-SNAPSHORT 2021-09-06 07:10
 * 类描述
 * 修订历史：
 * 日期：2021-09-06
 * 作者：generator
 * 参考：
 * 描述：AmsContractTransportPrice服务实现
 * @see null
 * @Copyright 湖南视拓信息技术股份有限公司. All rights reserved.
 */
@Service(value = "amsContractTruckPriceService")
public class AmsContractTruckPriceServiceImpl implements IAmsContractTruckPriceService {

    @Autowired
    private AmsContractTruckPriceMapper amsContractTruckPriceMapper;

    @Override
    public int addAmsContractTransportPrice(AmsContractTransportPrice amsContractTransportPrice) {
        long id;
        if(amsContractTruckPriceMapper.selectMaxId()==null){
            id = 1;
        }else {
            id = amsContractTruckPriceMapper.selectMaxId().longValue()+1;
        }
        //生成主键id
        amsContractTransportPrice.setPriceId(new BigDecimal(id));
        //获取路线Id和运价
        BigDecimal lineId = amsContractTransportPrice.getLineId();
        BigDecimal priceValue = amsContractTransportPrice.getPriceValue();
        //计算并插入吨公里
        BigDecimal priceTonKilometer = calculationPriceTonKilometer(lineId, priceValue);
        amsContractTransportPrice.setType(new BigDecimal(1));
        amsContractTransportPrice.setPriceTonKilometer(priceTonKilometer);
        amsContractTransportPrice.setInsertUsername("admin");
        amsContractTransportPrice.setInsertTime(new Date());
        amsContractTransportPrice.setUpdateUsername("admin");
        amsContractTransportPrice.setUpdateTime(new Date());
        amsContractTransportPrice.setInsertUpdateRemark("无");
        amsContractTransportPrice.setDeleted(new BigDecimal(0));
        return amsContractTruckPriceMapper.insertSelective(amsContractTransportPrice);
    }

    //计算吨公里值
    public BigDecimal calculationPriceTonKilometer(BigDecimal lineId,BigDecimal priceValue){
        //查询参考运距
        BigDecimal linePathLength = amsContractTruckPriceMapper.getLinePathLength(lineId);
        //计算吨公里	吨公里（元/吨.km） = 运价（元/吨）/ 参考运距（公里），保留4位小数
        BigDecimal priceTonKilometer = priceValue.divide(linePathLength, 4, BigDecimal.ROUND_HALF_UP);
        return priceTonKilometer;
    }

    @Override
    public List<Map<String, Object>> getAmsContractTransportPrice(Map<String, Object> map) {
        return amsContractTruckPriceMapper.getAmsContractTransportPrice(map);
    }

    @Override
    public int deleteAmsContractTransportPrice(BigDecimal priceId) {
        AmsContractTransportPrice amsContractTransportPrice = amsContractTruckPriceMapper.selectByPrimaryKey(priceId);
        amsContractTransportPrice.setDeleted(new BigDecimal(1));
        return amsContractTruckPriceMapper.updateByPrimaryKey(amsContractTransportPrice);
    }

    @Override
    public List<Map<String, Object>> getCarrierName() {
        return amsContractTruckPriceMapper.getCarrierName();
    }

    @Override
    public List<Map<String, Object>> getLineNo() {
        return amsContractTruckPriceMapper.getLineNo();
    }

    @Override
    public List<Map<String, Object>> getCapacityNumber() {
        return amsContractTruckPriceMapper.getCapacityNumber();
    }

    @Override
    public List<Map<String, Object>> getAddressDeliveryAddress(Map<String,Object> map) {
        return amsContractTruckPriceMapper.getAddressDeliveryAddress(map);
    }

    @Override
    public List<Map<String, Object>> selectByPriceId(BigDecimal priceId) {
        return amsContractTruckPriceMapper.selectByPriceId(priceId);
    }

    @Override
    public int updateAmsContractTransportPrice(AmsContractTransportPrice amsContractTransportPrice) {
        BigDecimal priceId = amsContractTransportPrice.getPriceId();
        //获取未修改前的汽运单价信息
        AmsContractTransportPrice amsContractTransportPrice1 = amsContractTruckPriceMapper.selectByPrimaryKey(priceId);
        //获取路线Id和运价
        BigDecimal lineId = amsContractTransportPrice.getLineId();
        BigDecimal priceValue = amsContractTransportPrice.getPriceValue();
        //若运价修改了
        if(amsContractTransportPrice.getPriceValue()!=amsContractTransportPrice1.getPriceValue()){
            //计算并插入吨公里
            BigDecimal priceTonKilometer = calculationPriceTonKilometer(lineId, priceValue);
            amsContractTransportPrice.setPriceTonKilometer(priceTonKilometer);
        }
        amsContractTransportPrice.setUpdateTime(new Date());
        return amsContractTruckPriceMapper.updateByPrimaryKeySelective(amsContractTransportPrice);
    }

    @Override
    public List<Map<String, Object>> getOriginAndNowOilPrice() {
        return amsContractTruckPriceMapper.getOriginAndNowOilPrice();
    }

    /**
     * 批量修改油价
     * @return
     */
    @Override
    public int batchUpdateTransportPriceByOilPrice(Map<String,Object> map) {
        //现执行价
        BigDecimal nowPriceValue;
        //油价变化率
        String n = map.get("n").toString();
        //承运商Id
        BigDecimal carrierId = (BigDecimal) map.get("carrierId");
        //批量修改焦炭运价
        List<Map<String, Object>> PriceValueList = amsContractTruckPriceMapper.getPriceValueList();
        //存储原运输单价实绩
        AmsContractTransportPrice oldAmsContractTransportPrice = new AmsContractTransportPrice();
        //存储新运输单价实绩
        AmsContractTransportPrice amsContractTransportPrice;
        //存储主键最大值
        BigDecimal i;
        //存储货物名称
        String cargonameValue = "";
        for (Map<String, Object> stringObjectMap : PriceValueList) {
            //查询货物名称
            cargonameValue = amsContractTruckPriceMapper.getCargonameValue(new BigDecimal(stringObjectMap.get("cargonameId").toString()));
            //获取原主键Id
            BigDecimal priceId = (BigDecimal) stringObjectMap.get("priceId");
            //获取承运商Id
            BigDecimal carrierId1 = (BigDecimal) stringObjectMap.get("carrierId");
            //修改原单价为历史数据
            oldAmsContractTransportPrice.setPriceId(priceId);
            oldAmsContractTransportPrice.setDeleted(new BigDecimal(1));
            amsContractTruckPriceMapper.updateByPrimaryKeySelective(oldAmsContractTransportPrice);
            //获取原单价的所有数据
            amsContractTransportPrice = amsContractTruckPriceMapper.selectByPrimaryKey(priceId);
            i = amsContractTruckPriceMapper.selectMaxId();
            //主键
            amsContractTransportPrice.setPriceId(i);
            //执行日期
            amsContractTransportPrice.setPriceDate(new Date());
            //修改现单价为执行数据
            amsContractTransportPrice.setDeleted(new BigDecimal(0));
            //若货物名称为焦炭
            if(cargonameValue.equals("焦炭")&&carrierId.intValue()==carrierId1.intValue()){
                //现执行价＝((63.00+63.00*0.13*n)+(66.50+66.50*0.3*n)+40.00)
                //调用焦炭运价计算公式计算现执行价
                try {
                    nowPriceValue = UpdateCokePriceValueByOilPrice(n);
                    //调整焦炭运价
                    amsContractTransportPrice.setPriceValue(nowPriceValue);
                    amsContractTruckPriceMapper.insert(amsContractTransportPrice);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if(cargonameValue.equals("纯苯")&&carrierId.intValue()==carrierId1.intValue()){
                //现执行价=(中标价*0.7)+(中标价*0.3)*(1+n)
                //调用纯苯运价计算公式计算现执行价
                try {
                    nowPriceValue = UpdatePureBenzenePriceValueByOilPrice(n,amsContractTransportPrice.getPriceValue());
                    amsContractTransportPrice.setPriceValue(nowPriceValue);
                    amsContractTruckPriceMapper.insert(amsContractTransportPrice);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }

//        if(updateOilPrice>0.05||updateOilPrice<-0.05) {
//            //逐条修改吨公里数
//            for (Map<String, Object> stringObjectMap : priceValueAndPriceTonKilometer) {
//                //主键Id
//                BigDecimal priceId = (BigDecimal) stringObjectMap.get("priceId");
//                //原运价
//                BigDecimal priceValue = (BigDecimal) stringObjectMap.get("priceValue");
//                //吨公里数
//                BigDecimal priceTonKilometer = (BigDecimal) stringObjectMap.get("priceTonKilometer");
//                //参考距离,四舍五入，保留两位小数
//                BigDecimal distance;
//                if (priceTonKilometer.doubleValue() != 0) {
//                    distance = priceValue.divide(priceTonKilometer, 2, BigDecimal.ROUND_HALF_UP);
//                    //现运价
//                    Double nowPriceValue = priceValue.doubleValue() * 0.65 + priceValue.doubleValue() * 0.35 * (1 + updateOilPrice);
//                    BigDecimal nowPriceValue1 = new BigDecimal(nowPriceValue);
//                    //新吨公里数
//                    BigDecimal newPriceTonKilometer = nowPriceValue1.divide(distance, 4, BigDecimal.ROUND_HALF_UP);
//                    //通过主键修改一条吨公里数
//                    amsContractTruckPriceMapper.updatePriceTonKilometer(priceId, newPriceTonKilometer);
//                }
//                return amsContractTruckPriceMapper.batchUpdateTransportPriceByOilPrice(updateOilPrice);
//            }
//        }
        return 1;
    }

    /**
     * 替换焦炭运价公式并进行计算
     * @param n
     * @return
     * @throws Exception
     */
    public BigDecimal UpdateCokePriceValueByOilPrice(String n) throws Exception {
        // 得到运算公式
        String formula_string = null;
            // 得到焦炭运价计算公式
            formula_string = amsContractTruckPriceMapper.getTruckFormula(new BigDecimal(12));
            // 替换焦炭运价计算公式
            formula_string = formula_string.replace("油价变动率",n);

        String s = toSufExpr(formula_string);
        return calSufExpr(s);
    }

    /**
     * 替换纯苯运价公式并进行计算
     * @param n
     * @return
     * @throws Exception
     */
    public BigDecimal UpdatePureBenzenePriceValueByOilPrice(String n,BigDecimal priceValue) throws Exception {
        // 得到运算公式
        String formula_string = null;
        // 得到纯苯运价计算公式
        formula_string = amsContractTruckPriceMapper.getTruckFormula(new BigDecimal(13));
        // 替换纯苯运价计算公式
        formula_string = formula_string.replace("中标价",priceValue.toString())
        .replace("油价变动率",n);
        String s = toSufExpr(formula_string);
        return calSufExpr(s);
    }

    /**
     * 中缀转换后缀表达式
     */
    public  String toSufExpr(String s) {
        StringBuffer sufExpr = new StringBuffer();
        Stack<String> stack = new Stack<>();
        //压入空方便判断
        stack.push(null);
        Pattern pattern = Pattern.compile("\\(int\\)|(?<!\\d|[)])-?\\d+(\\.\\d+)?|!=|>=|<=|==|[+\\-*/%()><=|]");
        Matcher matcher = pattern.matcher(s);
        while (matcher.find()) {
            String group = matcher.group();
            if (group.matches("!=|>=|<=|==|[+\\-*/%()><=|]|\\(int\\)")) {//是运算符
                if (group.equals("(")) {//之间压栈
                    stack.push(group);
                } else if (group.equals(")")) {//等于右括号，一直弹栈，直到遇到左括号
                    String pop = null;
                    while (!(pop = stack.pop()).equals("(")) {//遍历循环至左括号
                        sufExpr.append(pop + " ");
                    }
                } else {//是运算符：加减乘除强转，根据优先级压栈
                    while (getPriority(group) <= getPriority(stack.peek())) {
                        String pop = stack.pop();
                        System.out.println(pop);
                        System.out.println(sufExpr);
                        if(pop!=null&&!pop.equals("|")){
                            sufExpr.append(pop + " ");
                        }
                    }
                    stack.push(group);
                }
            } else {//是数字
                sufExpr.append(group + " ");
            }
        }
        String pop = null;
        while (null != (pop = stack.pop())) {
            sufExpr.append(pop + " ");
        }
        return sufExpr.toString();
    }

    /**
     * 设置出栈优先级，(：优先级最低，表示只有遇到)才能出栈
     * 当栈顶遇到优先级比自己小或等于的弹栈
     */
    public  Integer getPriority(String c) {
        if (c == null) {
            return 0;
        }
        switch (c) {
            case "(":
                return 1;
            case "+":
            case "-":
                return 2;
            case "*":
            case "/":
            case "%":
                return 3;
            case ">":
            case "<":
            case ">=":
            case "<=":
            case "!=":
            case "==":
                return 4;
            case "|":
                return 5;
            case "=":
                return 6;
            case "(int)":
                return 7;
        }
        return 0;
    }

    /**
     * 计算值
     */
    public  BigDecimal cal(BigDecimal a1, BigDecimal a2, String operator) throws Exception {
        switch (operator){
            case "+":return a1.add(a2);
            case "-":return a1.subtract(a2);
            case "*":return a1.multiply(a2);
            case "/":return a1.divide(a2,3,BigDecimal.ROUND_CEILING);
            case "%":return a1.divideAndRemainder(a2)[1];
            default:
                break;
        }
        throw new Exception("非法运算符");
    }

    public  Boolean calBoolean(BigDecimal a1, BigDecimal a2, String operator) throws Exception {
        switch (operator){
            case ">":return a1.compareTo(a2) == 1;
            case "<":return a1.compareTo(a2) == -1;
            case ">=":return a1.compareTo(a2) > -1;
            case "<=":return a1.compareTo(a2) < 1;
            case "!=":return a1.compareTo(a2) != 0;
            case "==":return a1.compareTo(a2) == 0;
            default:
                break;
        }
        throw new Exception("非法运算符");
    }

    /**
     * 计算后缀表达式
     */
    public  BigDecimal calSufExpr(String sufExpr) throws Exception {
        Stack<BigDecimal> stack = new Stack<>();
        Pattern pattern=Pattern.compile("\\(int\\)|-?\\d+(\\.\\d+)?|!=|>=|<=|==|[+\\-*/%><=]");//提取运算符和数字
        Matcher matcher = pattern.matcher(sufExpr);
        BigDecimal result=new BigDecimal("0.0");
        while (matcher.find()){
            String group = matcher.group();
            if(group.matches("[+\\-*/%]")){
                BigDecimal a1 =stack.pop();
                BigDecimal a2 =stack.pop();
                result = cal(a2, a1, group);
                stack.push(result);
            }
            else if(group.equals("(int)")){
                BigDecimal pop = stack.pop();
                BigDecimal bigDecimal = pop.setScale(0, BigDecimal.ROUND_DOWN);
                stack.push(bigDecimal);
            }
            else if(group.matches("!=|>=|<=|==|[><]")){
                BigDecimal a1 =stack.pop();
                BigDecimal a2 =stack.pop();
                stack.push(calBoolean(a2,a1,group)?new BigDecimal(1):new BigDecimal(0));
            }
            else if(group.equals("=")){
                BigDecimal a1 =stack.pop();
                BigDecimal a2 =stack.pop();
                if (a2.compareTo(new BigDecimal(1))==0) {
                    stack.push(a1);
                }
            }
            else {//为数字直接压栈
                stack.push(new BigDecimal(group));
            }
        }
        return result;
    }

}
